package org.kantega.jexmec.simple;

/*
 * Copyright 2010 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.*;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.*;

/**
 * Default implementation of PluginManager. Tracks the discovery of ClassLoaders and uses the provided
 * {@link org.kantega.jexmec.PluginLoader} implementations and a {@link org.kantega.jexmec.ServiceLocator} to load plugins from the ClassLoaders
 */
public class DefaultPluginManager<P extends Plugin> implements PluginManager<P>, PluginClassLoaderRegistry {

    private final Collection<PluginLoader<P>> pluginLoaders = new ArrayList<PluginLoader<P>>();

    private final Collection<PluginClassLoaderProvider> pluginClassLoaderProviders = new ArrayList<PluginClassLoaderProvider>();

    private final Map<ClassLoader, List<P>> dynamicPlugins = Collections.synchronizedMap(new HashMap<ClassLoader, List<P>>());

    private final ServiceLocator serviceLocator;
    private final Class<P> pluginClass;

    private enum State {STARTING, STARTED, STOPPING, STOPPED}
    private State state = State.STOPPED;

    private Map<P, ClassLoader> classLoaderMap = new HashMap<P, ClassLoader>();

    private Collection<PluginManagerListener<P>> listeners = Collections.synchronizedCollection(new ArrayList<PluginManagerListener<P>>());

    public DefaultPluginManager(Class<P> pluginClass) {
        this(pluginClass, new StaticServiceLocator());
    }

    public DefaultPluginManager(Class<P> pluginClass, ServiceLocator serviceLocator) {
        this(pluginClass, serviceLocator, null);
    }

    public DefaultPluginManager(Class<P> pluginClass, ServiceLocator serviceLocator, Class servicesInterface) {
        this.pluginClass = pluginClass;
        this.serviceLocator = new PluginManagerServiceLocatorWrapper(servicesInterface == null ? serviceLocator : new ServicesClassServiceLocator(serviceLocator, servicesInterface));
    }

    public DefaultPluginManager<P> addPluginClassLoaderProvider(PluginClassLoaderProvider pluginClassLoaderProvider) {
        if(state != State.STOPPED) {
            throw new IllegalStateException("Illegal state " + state +" when adding plugin class loader provider " + pluginClassLoaderProvider);
        }
        pluginClassLoaderProviders.add(pluginClassLoaderProvider);
        return this;
    }

    public DefaultPluginManager<P> addPluginLoader(PluginLoader<P> pluginLoader) {
        if(state != State.STOPPED) {
            throw new IllegalStateException("Illegal state " + state +" when adding plugin loader " +pluginLoader);
        }
        pluginLoaders.add(pluginLoader);
        return this;
    }


    public DefaultPluginManager<P> start() {
        if(state != State.STOPPED) {
            throw new IllegalStateException("Illegal state " + state +" when starting, expected " +State.STOPPED);
        }
        state = State.STARTING;

        checkDefaultPluginLoader();

        checkDefaultClassLoaderProvider();

        for(PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.start();
        }

        for(PluginClassLoaderProvider provider : pluginClassLoaderProviders) {
            provider.start(this);
        }

        state = State.STARTED;
        firePluginManagerStarted();
        return this;
    }

    public DefaultPluginManager<P> stop() {
        if(state != State.STARTED) {
            throw new IllegalStateException("Illegal state " + state +" when starting, expected " +State.STARTED);
        }
        state = State.STOPPING;

        for(PluginClassLoaderProvider provider : pluginClassLoaderProviders) {
            provider.stop();
        }
        
        for(PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.stop();
        }

        state = State.STOPPED;
        return this;
    }

    private void checkDefaultClassLoaderProvider() {
        if(pluginClassLoaderProviders.size() == 0) {
            pluginClassLoaderProviders.add(new StaticPluginClassLoaderProvider(getClass().getClassLoader()));
        }
    }

    private void checkDefaultPluginLoader() {
        if(pluginLoaders.size() == 0) {
            pluginLoaders.add(new SimpleConstructorPluginLoader<P>());
        }
    }

    private void firePluginManagerStarted() {
        for(PluginManagerListener listener : listeners) {
            listener.pluginManagerStarted();
        }
    }

    public synchronized List<P> getPlugins() {
        if(state != State.STARTED) {
            throw new IllegalStateException(DefaultPluginManager.class.getSimpleName() +" is not yet started. You need to call start() before calling getPlugins()");
        }
        List<P> plugins = new ArrayList<P>();
        for(List<P> loaderPlugins : dynamicPlugins.values()) {
            plugins.addAll(loaderPlugins);
        }
        return plugins;
    }


    public synchronized void addClassLoader(ClassLoader loader) {
        for(PluginManagerListener listener : listeners) {
            listener.classLoaderAdded(loader);
        }
        final List<P> loadedPlugins = loadPluginsFromClassLoader(serviceLocator, loader);
        for(P plugin : loadedPlugins) {
            classLoaderMap.put(plugin, loader);
        }

        dynamicPlugins.put(loader, loadedPlugins);
    }

    public synchronized void removeClassLoader(ClassLoader loader) {




        final List<P> pluginsRemoved = dynamicPlugins.get(loader);

        for(P plugin : pluginsRemoved) {
            classLoaderMap.remove(plugin);
        }
        
        dynamicPlugins.remove(loader);

        for(PluginManagerListener listener : listeners) {
            listener.classLoaderRemoved(loader);
        }
        unloadPluginsFromClassLoader(loader);

    }

    private void unloadPluginsFromClassLoader(ClassLoader loader) {
        for(PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.unloadPlugins(loader);
        }
    }

    private List<P> loadPluginsFromClassLoader(ServiceLocator serviceLocator, ClassLoader classLoader) {
        List<P> plugins = new ArrayList<P>();

        for(PluginLoader<P> pluginLoader : pluginLoaders) {
            plugins.addAll(pluginLoader.loadPlugins(pluginClass, classLoader, serviceLocator));
        }
        return plugins;
    }

    public void addPluginManagerListener(PluginManagerListener<P> pluginManagerListener) {
        listeners.add(pluginManagerListener);
    }

    public void removePluginManagerListener(PluginManagerListener<P> pluginManagerListener) {
        listeners.remove(pluginManagerListener);
    }

    private class ServicesClassServiceLocator<S> extends AbstractServiceLocator {
        private final ServiceLocator serviceLocator;
        private S servicesProxy;
        private final Class<S> serviceClass;


        public ServicesClassServiceLocator(ServiceLocator wrappedServiceLocator, Class<S> serviceClass) {
            super(serviceClass);
            this.serviceLocator = wrappedServiceLocator;
            this.servicesProxy = createServicesProxy(wrappedServiceLocator, serviceClass);
            this.serviceClass = serviceClass;
        }

        public <T> T lookupService(ServiceKey<T> serviceKey) {
            if(serviceKey.getServiceType() ==  this.serviceClass) {
                return serviceKey.getServiceType().cast(servicesProxy);
            } else {
                return serviceLocator.lookupService(serviceKey);
            }
        }

        public Set<ServiceKey> getServiceKeys() {
            Set<ServiceKey> serviceKeys = new HashSet<ServiceKey>();
            serviceKeys.add(ServiceKey.by(serviceClass));
            serviceKeys.addAll(super.getServiceKeys());
            serviceKeys.addAll(serviceLocator.getServiceKeys());
            return serviceKeys;
        }

        private S createServicesProxy(final ServiceLocator serviceLocator, final Class<S> servicesClass) {
            return servicesClass.cast(Proxy.newProxyInstance(getClass().getClassLoader(),
                    new Class[] {servicesClass},
                    new ServiceLookupInvocationHandler(serviceLocator)));
        }

    }

    private static class ServiceLookupInvocationHandler implements InvocationHandler {
            private ServiceLocator serviceLocator;

            public ServiceLookupInvocationHandler(ServiceLocator serviceLocator) {
                this.serviceLocator = serviceLocator;
            }

            public Object invoke(Object o, Method method, Object[] objects) throws Throwable {
                if(objects == null || objects.length == 0 ) {
                    return serviceLocator.lookupService(ServiceKey.by(method.getReturnType()));
                } else {
                    return serviceLocator.lookupService(ServiceKey.by(method.getReturnType(), ((Enum)objects[0]).name()));
                }
            }
        }

    private class PluginManagerServiceLocatorWrapper implements ServiceLocator {

        private ServiceLocator serviceLocator;

        public PluginManagerServiceLocatorWrapper(ServiceLocator serviceLocator) {

            this.serviceLocator = serviceLocator;
        }

        public <T> T lookupService(ServiceKey<T> serviceKey) {
            if(serviceKey.getServiceType() == PluginManager.class) {
                return serviceKey.getServiceType().cast(DefaultPluginManager.this);
            } else {
                return serviceLocator.lookupService(serviceKey);
            }
        }

        public Set<ServiceKey> getServiceKeys() {
            Set<ServiceKey> serviceKeys = new HashSet<ServiceKey>();
            serviceKeys.add(ServiceKey.by(PluginManager.class));
            serviceKeys.addAll(serviceLocator.getServiceKeys());
            return serviceKeys;
        }
    }

    public ClassLoader getClassLoader(P plugin) {
        return classLoaderMap.get(plugin);
    }
}